# -*- coding: utf-8 -*-
""" Scrutinizer: Cmd """

import os
from distutils import log

from . import Scrutinizer, ScrutinizerConfigError, ScrutinizerRunError


def validate_configuration(cfg):
    """ Validate configuration. """


class CmdScrutinizer(Scrutinizer):
    """ Cmd scrutinizer
        An example of how to implement a scrutinizer.
    """

    name = 'Command'

    description = "Execute an individual command"

    user_options = [
        ("cmd=", None, "Run this command"),
        ("time", None, "Measure execution time"),
    ]


    def initialize_options(self):
        """ Initialize object, really.
            Create the attributes and apply defaults, when available.
        """
        self.time = False
        self.cmd = None


    def finalize_options(self):
        """ Apply the configuration and prepare for running. """

        self.time = False
        log.debug("config: %s", self.config)
        self.cmd = self.config.get('cmd', None)


    def run(self):
        """ Run """

        if self.verbose:
            log.debug("Running command: '%s'", self.cmd)
        res = os.system(self.cmd)
        if self.verbose:
            log.debug("Command returned (code): '%d'", res)



""" Base for Scrutinizers """
import logging

class ScrutinizerError(Exception):
    """ The root of all Scrutinizer evil. """
    pass


class ScrutinizerConfigError(ScrutinizerError):
    """ Error in config. """
    pass


class ScrutinizerRunError(ScrutinizerError):
    """ Error during running. """
    pass


class Scrutinizer:
    """ Scrutinizer """

    # Subclasses must define:
    #   initialize_options()
    #     provide default values for all options; may be customized by
    #     setup script, by options from config file(s), or by command-line
    #     options
    #   finalize_options()
    #     decide on the final values for all options; this is called
    #     after all possible intervention from the outside world
    #     (command-line, option file, etc.) has been processed
    #   run()
    #     run the command: do whatever it is we're here to do,
    #     controlled by the command's various option values

    name = 'Base'


    def __init__(self, command):
    # def __init__(self):
        """Create and initialize a new Command object.  Most importantly,
        invokes the 'initialize_options()' method, which is the real
        initializer and depends on the actual command being
        instantiated.
        """
        # logging.error("301")
        # late import because of mutual dependence between these classes
        # from distutils.dist import Distribution
        from setuptools_scrutinize.setuptools_command import Scrutinize

        # logging.error("302")
        if not isinstance(command, Scrutinize):
            raise TypeError("command must be a Scrutinize instance")
        # logging.error("303")
        # if not isinstance(dist, Distribution):
        #     raise TypeError("dist must be a Distribution instance")
        # if self.__class__ is Command:
        #     raise RuntimeError("Command is an abstract class")
        if self.__class__ is Scrutinizer:
            raise RuntimeError("Scrutinizer is an abstract class")

        # self.distribution = command.distrib
        self.command = command
        self.initialize_options()

        self._dry_run = None

        # self.verbose = dist.verbose

        # Some commands define a 'self.force' option to ignore file
        # timestamps, but methods defined *here* assume that
        # 'self.force' exists for all commands.  So define it here
        # just to be safe.
        self.force = None

        # The 'help' flag is just used for command-line parsing, so
        # none of that complicated bureaucracy is needed.
        self.help = 0

        # 'finalized' records whether or not 'finalize_options()' has been
        # called.  'finalize_options()' itself should not pay attention to
        # this flag: it is the business of 'ensure_finalized()', which
        # always calls 'finalize_options()', to respect/update it.
        self.finalized = 0

        self.config = None


    def initialize_options(self):
        """Set default values for all the options that this command
        supports.  Note that these defaults may be overridden by other
        commands, by the setup script, by config files, or by the
        command-line.  Thus, this is not the place to code dependencies
        between options; generally, 'initialize_options()' implementations
        are just a bunch of "self.foo = None" assignments.

        This method must be implemented by all command classes.
        """
        raise RuntimeError("abstract method -- subclass %s must override"
                           % self.__class__)


    def finalize_options(self):
        """Set final values for all the options that this command supports.
        This is always called as late as possible, ie.  after any option
        assignments from the command-line or from other commands have been
        done.  Thus, this is the place to code option dependencies: if
        'foo' depends on 'bar', then it is safe to set 'foo' from 'bar' as
        long as 'foo' still has the same value it was assigned in
        'initialize_options()'.

        This method must be implemented by all command classes.
        """
        raise RuntimeError("abstract method -- subclass %s must override"
                           % self.__class__)


    def run(self):
        """A command's raison d'etre: carry out the action it exists to
        perform, controlled by the options initialized in
        'initialize_options()', customized by other commands, the setup
        script, the command-line, and config files, and finalized in
        'finalize_options()'.  All terminal output and filesystem
        interaction should be done by 'run()'.

        This method must be implemented by all command classes.
        """
        raise RuntimeError("abstract method -- subclass %s must override"
                           % self.__class__)




# -*- coding: utf-8 -*-
""" Scrutinizer: Pylint """

from typing import Final, Tuple, Any
import os
import re
import sys
from distutils import log
from pylint import lint  # type: ignore
from pylint.__pkginfo__ import numversion as pylint_version  # type: ignore
from pkg_resources import *

from . import Scrutinizer
from setuptools_scrutinize import ScrutinizerConfigError, ScrutinizerRunError


_opts: Final[Tuple[str, Any]] = lint.Run.LinterClass.make_options()


def user_options():
    parsed = []

    for longopt, params in _opts:
        desc = params.get("help", None)

        parsed.append(("lint-" + longopt + "=", None, desc))
    return parsed


def validate_configuration(cfg):
    """ Validate configuration.
    """
    pass

class PylintScrutinizer(Scrutinizer):
    """ Pylint scrutinizer """

    name = 'Pylint'

    description = "Pylint code checker"

    user_options = user_options() + [
        (
            "lint-packages=",
            None,
            "Report on just these packages. These arguments are passed straight "
            "through to pylint as the module_or_package arguments. As such, they "
            "can be paths to files or packages",
        ),
        ("lint-exclude-packages=", None, "exclude these packages"),
        ("lint-output=", None, "output report into this file"),
        ("lint-rcfile=", None, "pylint configuration file"),
    ]


    def initialize_options(self):
        """ Initialize object, really.
            Create the attributes and apply defaults, when available.
        """
        self.lint_packages = ""
        self.lint_exclude_packages = "tests test"
        self.lint_output = None
        # self.lint_rcfile = self.distribution.scrutinize_rcfile
        self.lint_rcfile = None
        self.distribution = self.command.distribution
        # self.command = None
        for longopt, params in _opts:
            key = "lint_" + longopt.replace("-", "_").rstrip("=")
            setattr(self, key, None)


    def finalize_options(self):
        """ Apply the configuration and prepare for running.
        """
        self.lint_rcfile = self.config
        self.lint_packages = [package.strip() for package in re.split(r"[\s,]+", self.lint_packages) if package != ""]  # pylint disable=line-too-long
        self.lint_exclude_packages = [module.strip() for module in re.split(r"[\s,]+", self.lint_exclude_packages)]  # pylint disable=line-too-long
        if self.lint_output:
            out_dir = os.path.dirname(self.lint_output)
            if out_dir:
                if sys.version_info >= (3, 2):
                    os.makedirs(out_dir, exist_ok=True)
                elif not os.path.exists(out_dir):
                    os.makedirs(out_dir)
            self.lint_output = open(self.lint_output, "w")

    def with_project_on_sys_path(self, func, func_args, func_kwargs):
        """ Fix the project path to include the eggs.
            Now you include also the build dependencies which are
            installed locally.
            Supports both Python 2 and Python 3.
        """
        if sys.version_info >= (3,) and getattr(self.distribution, "use_2to3", False):
            # If we run 2to3 we can not do this inplace:

            # Ensure metadata is up-to-date
            self.command.reinitialize_command("build_py", inplace=0)
            self.command.run_command("build_py")
            bpy_cmd = self.command.get_finalized_command("build_py")
            build_path = normalize_path(bpy_cmd.build_lib)

            # Build extensions
            self.command.reinitialize_command("egg_info", egg_base=build_path)
            self.command.run_command("egg_info")

            self.command.reinitialize_command("build_ext", inplace=0)
            self.command.run_command("build_ext")
        else:
            # Without 2to3 inplace works fine:
            self.command.run_command("egg_info")

            # Build extensions in-place
            self.command.reinitialize_command("build_ext", inplace=1)
            self.command.run_command("build_ext")

        ei_cmd = self.command.get_finalized_command("egg_info")

        old_path = sys.path[:]
        old_modules = sys.modules.copy()

        try:
            sys.path.insert(0, normalize_path(ei_cmd.egg_base))
            working_set.__init__()
            add_activation_listener(lambda dist: dist.activate())
            require("%s==%s" % (ei_cmd.egg_name, ei_cmd.egg_version))
            if self.verbose:
                log.debug("Running with args: %s; %s", func_args, func_kwargs)
            return func(*func_args, **func_kwargs)
        finally:
            sys.path[:] = old_path
            sys.modules.clear()
            sys.modules.update(old_modules)
            working_set.__init__()

    def run(self):
        options = []
        for longopt, params in _opts + (("rcfile", None),):
            # value = getattr(self, "lint_" + longopt.replace("-", "_"))
            value = self.config.get("lint_" + longopt.replace("-", "_"))
            if value is not None:
                if " " in value:
                    value = '"' + value + '"'
                options.append("--{0}={1}".format(longopt, value))

        # if self.verbose:
        #     print("Config:")
        #     print(options)

        # # Make sure the dependencies are available
        # if self.distribution.install_requires:
        #     self.distribution.fetch_build_eggs(self.distribution.install_requires)
        #
        # if self.distribution.tests_require:
        #     self.distribution.fetch_build_eggs(self.distribution.tests_require)

        if self.lint_packages:
            # The user explicitly specified the paths/packages to send to lint
            files = self.lint_packages
        else:
            # With no packages specified, find all of them and pass them
            # through the filter
            base = self.command.get_finalized_command("build_py")
            files = [
                filename
                for (package, module, filename) in base.find_all_modules()
                if package not in self.lint_exclude_packages
            ]

        if self.lint_output:
            stdout, sys.stdout = sys.stdout, self.lint_output
            stderr, sys.stderr = sys.stderr, self.lint_output
        try:
            kwargs = {}
            if pylint_version < (2, 0, 0):
                kwargs.update({"exit": False})
            else:
                kwargs.update({"do_exit": False})
            lint_runner = self.with_project_on_sys_path(lint.Run, [options + files], kwargs)
            if lint_runner.linter.msg_status:
                raise ScrutinizerRunError("lint error %s." % lint_runner.linter.msg_status)
        finally:
            if self.lint_output:
                sys.stdout = stdout
                sys.stderr = stderr



from setuptools import setup

setup(
    entry_points={
        "distutils.commands": [
            "scrutinize = setuptools_scrutinize.setuptools_command:Scrutinize"
            ],
        "distutils.setup_keywords": [
            "scrutinize_config = setuptools_scrutinize.setuptools_command:validate_config_file_exists"
            ],
    },
)



# -*- coding: utf-8 -*-
""" Scrutinize setuptools command plugin. """

from typing import Final
import os
import sys
import setuptools
import configparser
import importlib.util
from distutils import log
from distutils.errors import DistutilsOptionError, DistutilsSetupError, DistutilsError

from setuptools_scrutinize import ScrutinizerConfigError, ScrutinizerRunError

SETUP_CONFIG_OPTION: Final[str] = 'scrutinize_config'
DEFAULT_CONFIG_FILEPATH: Final[str] = '.scrutinize'
CONFIG_SECTION: Final[str] = 'scrutinize'

def known_scrutinizers():
    return {
            'pylint': {
                'module': 'setuptools_scrutinize.pylint',
                'class': 'PylintScrutinizer',
                },
            'cmd': {
                'module': 'setuptools_scrutinize.cmd',
                'class': 'CmdScrutinizer',
                },
            }


def validate_config_file_exists(dist, attr, value):
    """ Validate the configuration file exists. """

    if not os.path.exists(value):
        raise DistutilsSetupError("Cannot find Scrutinize configuration file '%s'" % value)


class DistutilsCheckError(DistutilsError):
    pass


class Scrutinize(setuptools.Command):

    description = "Run scrutinizing quality assurance"

    user_options = [
            ('list', 'l', "List scrutinizers"),
            # ('lint-config', None, "Check that Scurinize configuration file is valid"),
            # ('lint-output=', None, "output report into this file"),
            ('config=', None, "Scrutinize configuration file, default: '" +
                                DEFAULT_CONFIG_FILEPATH +
                                "'. In setup.cfg/setup.py, use option '"  + SETUP_CONFIG_OPTION +
                                "'."),
            ]


    def initialize_options(self):
        self.list = None
        self.lint_config = None
        self.config = DEFAULT_CONFIG_FILEPATH if DEFAULT_CONFIG_FILEPATH else self.distribution.scrutinize_config
        self.scrutinizers = [ ]


    def finalize_options(self):
        cfg_file = self.config
        if not os.path.exists(cfg_file):
            raise DistutilsSetupError("Cannot find Scrutinize configuration file '%s'" % cfg_file)
        try:
            config = configparser.ConfigParser()
            config.read(cfg_file)
        except configparser.Error:
            raise DistutilsSetupError("Error in Scrutinize configuration file '%s': %s" % (cfg_file, sys.exc_info()[1]))

        # Get the global cfg_files, then remove the section
        # (so it doesn't count as a scrutinizer section later).
        packages = ''
        exclude_packages = 'tests test'
        if config.has_section(CONFIG_SECTION):
            for option in ['verbose', 'dry_run']:
                try:
                    if config.has_option(CONFIG_SECTION, option):
                        setattr(self, option, config.getboolean(CONFIG_SECTION, option))
                except ValueError:
                    raise DistutilsOptionError(f"Option {CONFIG_SECTION}.{option} in '{cfg_file}' is not boolean")

            # packages
            if config.has_option(CONFIG_SECTION, 'packages'):
                packages = config.get(CONFIG_SECTION, 'packages')
            if config.has_option(CONFIG_SECTION, 'exclude_packages'):
                exclude_packages = config.get(CONFIG_SECTION, 'exclude_packages')

            config.remove_section(CONFIG_SECTION)

        # Check that there is a class/module for each scrutinizer.
        for scrutinizer in config.sections():

            if config.has_option(CONFIG_SECTION, 'packages'):
                packages = config.get(CONFIG_SECTION, 'packages')
                config.remove_option(CONFIG_SECTION, 'packages')
            if config.has_option(CONFIG_SECTION, 'exclude_packages'):
                exclude_packages = config.get(CONFIG_SECTION, 'exclude_packages')
                config.remove_option(CONFIG_SECTION, 'exclude_packages')
            mod_cfg = config[scrutinizer]

            if scrutinizer not in known_scrutinizers():
                raise DistutilsOptionError("Scrutinizer '%s' not found" % scrutinizer)
            # TODO If the scrutinizer is an exe, check that it can be run.

            mod_name = known_scrutinizers().get(scrutinizer).get('module')
            class_name = known_scrutinizers().get(scrutinizer).get('class')
            try:
                spec = importlib.util.find_spec(mod_name)
                module = importlib.util.module_from_spec(spec)
                sys.modules[mod_name] = module
                spec.loader.exec_module(module)
                module.validate_configuration(mod_cfg)
                scru_class = getattr(module, class_name)
                scru = scru_class(self)  # N.B. We pass self, i.e. Command as param.

                # Configure for run
                scru.initialize_options()
                scru.verbose = self.verbose
                scru.dry_run = self.dry_run
                scru.config = mod_cfg
                scru.lint_packages = packages
                scru.lint_exclude_packages = exclude_packages
                scru.finalize_options()
                
                self.scrutinizers.append(scru)

            except ScrutinizerConfigError:
                err = sys.exc_info()[1]
                raise DistutilsOptionError(f"Error in config of Scrutinizer '{scrutinizer}' ({mod_name}): {err}")
            except ModuleNotFoundError:
                err = sys.exc_info()[1]
                raise DistutilsOptionError(f"Cannot load Scrutinizer '{scrutinizer}' ({mod_name}): {err}")


    def run(self):
        """ Run """

        # Make sure the dependencies are available
        if self.distribution.install_requires:
            self.distribution.fetch_build_eggs(self.distribution.install_requires)

        if self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)

        # if self.lint_packages:
        #     # The user explicitly specified the paths/packages to send to lint
        #     files = self.lint_packages
        # else:
        #     # With no packages specified, find all of them and pass them
        #     # through the filter
        #     base = self.command.get_finalized_command("build_py")
        #     files = [
        #         filename
        #         for (package, module, filename) in base.find_all_modules()
        #         if package not in self.lint_exclude_packages
        #     ]

        for scru in self.scrutinizers:
            if self.verbose:
                log.debug("Scrutinizer: %s", scru.name)
            try:
                scru.run()
            except ScrutinizerRunError:
                raise DistutilsError(f"Scrutinizer {scru.name} discovered errors")
